local MODNAME = ...

-------------------------------------------------------------------------------
-- Localized Lua globals
-------------------------------------------------------------------------------
local _G = getfenv(0)

local assert = _G.assert
local ceil = _G.ceil
local concat = _G.table.concat
local floor = _G.floor
local format = _G.format
local match = _G.match
local strfind = _G.strfind
local tonumber = _G.tonumber
local tostring = _G.tostring
local type = _G.type

local ipairs = _G.ipairs
local pairs = _G.pairs
local select = _G.select
local sort = _G.sort
local tinsert = _G.tinsert
local tremove = _G.tremove
local wipe = _G.table.wipe

-------------------------------------------------------------------------------
-- AddOn namespace
-------------------------------------------------------------------------------
local Syllabus = LibStub("AceAddon-3.0"):NewAddon(MODNAME, "AceEvent-3.0", "AceConsole-3.0")
_G[MODNAME] = Syllabus

local LDB = LibStub("LibDataBroker-1.1")
local QTip = LibStub("LibQTip-1.0")

local dev = false
--[===[@debug@
dev = true
--@end-debug@]===]

local alpha = false
--@alpha@
alpha = true
--@end-alpha@

local L = LibStub("AceLocale-3.0"):GetLocale(MODNAME, "enUS", true, dev)

local DataObj
local tooltip

local version = GetAddOnMetadata(MODNAME, "Version")
version = dev and (version .. " - Development") or (alpha and version .. " - Alpha") or version
Syllabus.version = version

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------
local MY_CLASS, MY_CLASS_UPPER = UnitClass("player")
local MY_LEVEL = UnitLevel("player")
local MY_NAME = UnitName("player")
local MY_REALM = GetRealmName()
local MY_RACE = UnitRace("player")
local MY_FACTION = UnitFactionGroup("player")

local COLOR_TABLE = CUSTOM_CLASS_COLORS or RAID_CLASS_COLORS
local CLASS_COLORS = {}

local MY_PROFESSIONS = {}
Syllabus.prof = MY_PROFESSIONS

for k, v in pairs(COLOR_TABLE) do
	CLASS_COLORS[k] = ("%2x%2x%2x"):format(v.r * 255, v.g * 255, v.b * 255)
end

-- Profession titles (this will need to be localized
-- The key is the max rank for a skill level, the number is the SpellID where we can get the title name
-- I'm trying to keep the localization needs to a minimum but I'm not sure it's such a good idea because of
-- the number of GetSpellInfo that are called on the first update
--	[_G.PROFESSIONS_ARCHAEOLOGY]		= { [0] = 80836, [75] = },
local TITLES = {
	[GetSpellInfo(2259)] = {	-- Alchemy
		[75] = 2275,
		[150] = 2280,
		[225] = 3465,
		[300] = 11612,
		[375] = 28597,
		[450] = 65281,
	},
	[GetSpellInfo(2018)] = {	-- Blacksmithing
		[75] = 2020,
		[150] = 2021,
		[225] = 3539,
		[300] = 9786,
		[375] = 29845,
		[450] = 65282,
	},
	[GetSpellInfo(7411)] = {	-- Enchanting
		[75] = 7414,
		[150] = 7415,
		[225] = 7416,
		[300] = 13921,
		[375] = 28030,
		[450] = 65285,
	},
	[GetSpellInfo(4036)] = {	-- Engineering
		[75] = 4039,
		[150] = 4040,
		[225] = 4041,
		[300] = 12657,
		[375] = 30351,
		[450] = 61464,
	},
	[GetSpellInfo(9134)] = {	-- Herbalism
		[75] = 2372,
		[150] = 2373,
		[225] = 3571,
		[300] = 11994,
		[375] = 28696,
		[450] = 65288,
	},
	[GetSpellInfo(45357)] = {	-- Inscription
		[75] = 45375,
		[150] = 45376,
		[225] = 45377,
		[300] = 45378,
		[375] = 45379,
		[450] = 45380,
	},
	[GetSpellInfo(25229)] = { 	-- Jewelcrafting
		[75] = 25245,
		[150] = 24246,
		[225] = 28896,
		[300] = 28899,
		[375] = 28901,
		[450] = 51310,
	},
	[GetSpellInfo(2108)] = {	-- Leatherworking
		[75] = 2155,
		[150] = 2154,
		[225] = 3812,
		[300] = 10663,
		[375] = 32550,
		[450] = 51301,
	},
	[GetSpellInfo(2575)] = {	-- Mining
		[75] = 2581,
		[150] = 2582,
		[225] = 3568,
		[300] = 10249,
		[375] = 29355,
		[450] = 50309,
	},
	[GetSpellInfo(8613)] = {	-- Skinning
		[75] = 8615,
		[150] = 8619,
		[225] = 8620,
		[300] = 10769,
		[375] = 32679,
		[450] = 50307,
	},
	[GetSpellInfo(3908)] = {	-- Tailoring
		[75] = 3911,
		[150] = 3912,
		[225] = 3913,
		[300] = 12181,
		[375] = 26791,
		[450] = 51308,
	},
	[_G.PROFESSIONS_COOKING] = {
		[75] = 2551,
		[150] = 3412,
		[225] = 54257,
		[300] = 18261,
		[375] = 54256,
		[450] = 65291,
	},
	[_G.PROFESSIONS_FIRST_AID] = {
		[75] = 3279,
		[150] = 3280,
		[225] = 54254,
		[300] = 10847,
		[375] = 54255,
		[450] = 65292,
	},
	[_G.PROFESSIONS_FISHING] = {
		[75] = 7733,
		[150] = 7734,
		[225] = 54083,
		[300] = 18249,
		[375] = 54084,
		[450] = 65293,
	},
}

-- Implied skills: knowing a skill imply that you know some other skills
-- The professions can probably be done that way for the TITLES
local IMPLIED_SKILLS = {
	--33388 -- Apprentice Riding
	--33391 -- Journeyman Riding
	--34090 -- Expert Riding
	--34091 -- Artisan Riding
	--54197 -- Cold Weather Flying
	--90265 -- Master Riding
	[33391] = { 33388 },
	[34090] = { 33391, 33388 },
	[34091] = { 34090, 33391, 33388 },
	[54197] = { 34090, 33391, 33388 },
	[90265] = { 34090, 33391, 33388 },
	[86525] = { 750 }, -- Plate Specialization => Plate Mail
}

-- Abilities that are requirements for skills
local IS_ABILITY = {
	[71] 		= true, -- Defensive Stance
	[339] 	= true, -- Entangling Roots
	[527] 	= true, -- Dispel Magic
	[768] 	= true, -- Cat Form
	[1120] 	= true, -- Drain Soul
	[1515] 	= true, -- Tame Beast
	[1784] 	= true, -- Stealth
	[2458] 	= true, -- Berserker Stance
	[2836] 	= true, -- Detect Traps
	[5487] 	= true, -- Bear Form
	[20219] 	= true, -- Gnomish Engineer
	[20222] 	= true, -- Goblin Engineer
	[33391] 	= true, -- Journeyman Riding
	[34091] 	= true, -- Artisan Riding
	[42955] 	= true, -- Conjure Refreshment
}

-- Will be populated during initialization
local ABILITY_NAME = {} -- SpellID => Long Ability Name
local SHORT_ABILITY_NAME = {} -- Long Ability Name => Short Ability Name

-- Skill that requires these ability will not be displayed unless the ability is known
-- Mostly, if you are not a Gnome Engineer, you don't see that you could learn Gnomish Shrink Ray and
-- the like if you were a Gnomish Engineer. Speciality just generated to much crap.
-- If someone has an idea for a better idea, let me know - LaoTseu
local IS_ABILITY_EXCEPTION = {
	[20219] 	= true, -- Gnomish Engineer
	[20222] 	= true, -- Goblin Engineer
}
local IS_ABILITY_EXCEPTION_NAME = {} -- Long Ability Name => SpellID, need it to test against a name
-- Debug
if dev then
	Syllabus.ABILITY_NAME = ABILITY_NAME
	Syllabus.SHORT_ABILITY_NAME = SHORT_ABILITY_NAME
	Syllabus.IS_ABILITY_EXCEPTION_NAME = IS_ABILITY_EXCEPTION_NAME
end


-- To find the spell_id using the texture
local TRACKING = {
	DRUID = {
		["Interface\\Icons\\Ability_Tracking"] = 5225, -- Track Humanoids
	},
	HUNTER = {
		["Interface\\Icons\\Ability_Racial_Avatar"] = 19882, -- Track Giants
		["Interface\\Icons\\INV_Misc_Head_Dragon_01"] = 19879, -- Track Dragonkin
		["Interface\\Icons\\Spell_Holy_PrayerOfHealing"] = 19883, -- Track Humanoids
		["Interface\\Icons\\Spell_Shadow_SummonFelHunter"] = 19878, -- Track Demons
		["Interface\\Icons\\Ability_Stealth"] = 19885, -- Track Hidden
		["Interface\\Icons\\Spell_Shadow_DarkSummoning"] = 19884, -- Track Undead
		["Interface\\Icons\\Ability_Tracking"] = 1494, -- Track Beasts
		["Interface\\Icons\\Spell_Frost_SummonWaterElemental"] = 19880, -- Track Elementals
	},
}

-- For debug
--Syllabus.TITLES = TITLES

-- Special case for skills that require for the crafter to be able to wear a certain type of armor
local NEED_ARMOR = {
	[41320]		= 'Cloth', -- Destruction Holo-gogs
	[41321]		= 'Cloth', -- Powerheal 4000 Lens
	[56484]		= 'Cloth', -- Visage Liquification Goggles
	[41317]		= 'Leather', -- Deathblow X11 Goggles
	[41319]		= 'Leather', -- Magnified Moon Specs
	[56481]		= 'Leather', -- Weakness Spectralizers
	[56486]		= 'Leather', -- Greensight Gogs
	[41314]		= 'Mail', -- Surestrike Goggles v2.0
	[41315]		= 'Mail', -- Gadgetstorm Goggles
	[41316]		= 'Mail', -- Living Replicator Specs
	[56487]		= 'Mail', -- Electroflux Sight Enhancers
	[56574]		= 'Mail', -- Truesight Ice Blinders
	[40274]		= 'Plate',	-- Furious Gizmatic Goggles
	[41311]		= 'Plate', -- Justicebringer 2000 Specs
	[41312]		= 'Plate', -- Tankatronic Goggles
	[56480]		= 'Plate', -- Armored Titanium Goggles
	[56483]		= 'Plate', -- Charged Titanium Specs
	[62271]		= 'Plate', -- Unbreakable Healing Amplifiers
}

-- Type of armor that a class can wear at max level
local ARMOR_BY_CLASS = {
	MAGE			= 'Cloth',
	PRIEST		= 'Cloth',
	WARLOCK		= 'Cloth',
	DRUID			= 'Leather',
	ROGUE			= 'Leather',
	HUNTER		= 'Mail',
	SHAMAN		= 'Mail',
	PALADIN		= 'Plate',
	WARRIOR		= 'Plate',
	DEATHKNIGHT	= 'Plate',
}

-- Shared skills - skills that are shared between all classes even though
-- IsTradeskillTrainer() return false
local SHARED_SKILLS = {
	[33388]	= true, -- Apprentice Riding
	[33391]	= true, -- Journeyman Riding
	[34090]	= true, -- Expert Riding
	[34091]	= true, -- Artisan Riding
	[54197]	= true, -- Cold Weather Flying
	[90265]	= true, -- Master Riding
}


-------------------------------------------------------------------------------
-- Variables
-------------------------------------------------------------------------------
local update_queued = false
local in_combat = false

local db
local faction_db
local known_db

-------------------------------------------------------------------------------
-- Functions
-------------------------------------------------------------------------------

-- print functions
function Syllabus:Print(msg)
	_G.SELECTED_CHAT_FRAME:AddMessage(MODNAME .. ": " .. msg, 0.5, 1, 0)
end

function Syllabus:debug(msg, ...)
	if dev then
		self:err(msg, ...)
	end
end

function Syllabus:err(msg, ...)
	_G.geterrorhandler()(msg:format(...) .. " - " .. _G.time())
end

local function print(...) Syllabus:Print(...) end
local function debug(...) Syllabus:debug(...) end
local function err(...) Syllabus:err(...) end

-- Table copy function -- borrowed from AceDB-3.0.lua
local function copyTable(src, dest)
	if type(dest) ~= "table" then dest = {} end
	if type(src) == "table" then
		for k,v in pairs(src) do
			if type(v) == "table" then
				-- try to index the key first so that the metatable creates the defaults, if set, and use that table
				v = copyTable(v, dest[k])
			end
			dest[k] = v
		end
	end
	return dest
end

-- Table reuse functions
local AcquireTable, ReleaseTable
do
	local table_cache = {}

	-- Returns a table
	function AcquireTable()
		local tbl = tremove(table_cache) or {}
		return tbl
	end

	-- Cleans the table and stores it in the cache
	function ReleaseTable(tbl)
		if not tbl then return end

		-- Nested tables ?
		for i=1,#tbl do
			if type(tbl[i]) == "table" then ReleaseTable(tbl[i]) end
		end

		wipe(tbl)
		tinsert(table_cache, tbl)
	end
end	-- do block

-- Find the base cost from the real cost and the reputation standing
local function GetBaseCost(cost, StandingId)
	assert(StandingId and StandingId > 3,"Invalid reputation StandingId (must be between 4 and 8): "..tostring(StandingId))

	local rebate = 1 - (StandingId - 4) * .05
	return floor(cost / rebate)
end

local function GetRealCost(cost, StandingId)
	assert(StandingId and StandingId > 3,"Invalid reputation StandingId (must be between 4 and 8): "..tostring(StandingId))

	local rebate = 1 - (StandingId - 4) * .05
	return ceil(cost * rebate)
end

local GetReputationStanding
local GetCheapestCost
do -- Closure for the local tables

	local rep_index_cache = {}
	function GetReputationStanding(rep_name)

		if rep_index_cache[rep_name] then
			local name, _, standingId = _G.GetFactionInfo(rep_index_cache[rep_name])

			if name == rep_name then
				return standingId
			else
				-- Reputations have change, the index is no longer valid
				rep_index_cache[rep_name] = nil
			end
		end

		for factionIndex = 1, _G.GetNumFactions() do
			local name, _, standingId, _, _, _, _, _, isHeader = _G.GetFactionInfo(factionIndex)
			if not isHeader and rep_name == name then
				rep_index_cache[rep_name] = factionIndex
				return standingId
			end
		end

		return 0
	end

	local rep_standings  = {}
	local best_rep = {}
	function GetCheapestCost(cost,rep_table)
		--err("cost=%s, rep_table=%s",tostring(cost),tostring(rep_table))
		if rep_table == nil then return cost, "" end
		local best_standing = 0

		-- Refesh the rep_standings
		for factionIndex = 1, _G.GetNumFactions() do
			local rep_name, _, standingId, _, _, _, _, _, isHeader = _G.GetFactionInfo(factionIndex)
			if not isHeader then
				-- We refresh the index cache while were's at it
				rep_index_cache[rep_name] = factionIndex
				rep_standings[rep_name] = standingId
			end
		end

		-- Find the best standing
		for rep_name in pairs(rep_table) do
			if rep_standings[rep_name] and rep_standings[rep_name] > best_standing then
				best_standing = rep_standings[rep_name]
			end
		end

		if best_standing == 0 then return cost, "" end

		-- Build a string of the best Reputation factions
		wipe(best_rep)
		for rep_name in pairs(rep_table) do
			if rep_standings[rep_name] == best_standing then
				tinsert(best_rep,rep_name)
			end
		end
		sort(best_rep)

		return GetRealCost(cost, best_standing), concat(best_rep, ", ")

	end

end -- End of closure

-- Scan the current class spells to see what the toon already know
function Syllabus:InitKnownspells()
	-- Find ability names
	for spell_id in pairs(IS_ABILITY) do
		local name, type = _G.GetSpellInfo(spell_id)
		-- In English, the type is put between () in the requires text from the skills,
		-- I hope it's the same in the other locales
		local full_name = (type and type ~= "" and ("%s (%s)"):format(name, type)) or name
		ABILITY_NAME[spell_id] = full_name
		SHORT_ABILITY_NAME[full_name] = name
	end

	for spell_id in pairs(IS_ABILITY_EXCEPTION) do
		local name, type = _G.GetSpellInfo(spell_id)
		local full_name = (type and type ~= "" and ("%s (%s)"):format(name, type)) or name
		IS_ABILITY_EXCEPTION_NAME[full_name] = spell_id
	end

	-- Scan the Spellbook
	local i = 1
	local type, spell_id = _G.GetSpellBookItemInfo(i, BOOKTYPE_SPELL)
	while type ~= nil do
		--err("type=%s,spell_id=%s",tostring(type),tostring(spell_id))
		if type == "SPELL" or type == "FLYOUT" then
			local skill_name = (spell_id and _G.GetSpellInfo(spell_id)) or nil
			if skill_name then
				known_db[MY_CLASS_UPPER..":"..skill_name] = spell_id

				-- Abilities
				if IS_ABILITY[spell_id] then
					known_db["ABILITY:"..ABILITY_NAME[spell_id]] = spell_id
				end
			end

			-- Implied skills
			local implied = spell_id and IMPLIED_SKILLS[spell_id]
			if implied then
				for _, spell_id in ipairs(implied) do
					known_db[MY_CLASS_UPPER..":".._G.GetSpellInfo(spell_id)] = spell_id

					-- Abilities
					if IS_ABILITY[spell_id] then
						known_db["ABILITY:"..ABILITY_NAME[spell_id]] = spell_id
					end
				end
			end
		end
		i = i + 1
		type, spell_id = _G.GetSpellBookItemInfo(i, BOOKTYPE_SPELL)
	end

	-- Scan the mount
	for i = 1, GetNumCompanions("MOUNT") do
		local spell_id = select(3,GetCompanionInfo("MOUNT", i))
		known_db[MY_CLASS_UPPER..":".._G.GetSpellInfo(spell_id)] = spell_id
	end

	-- Scan the trackings
	if TRACKING[MY_CLASS_UPPER] then
		for i = 1, GetNumTrackingTypes() do
			-- Yes, the only way to find the spell without knowing the name is by the texture...
			local name, texture = GetTrackingInfo(i)
			known_db[MY_CLASS_UPPER..":"..name] = TRACKING[MY_CLASS_UPPER][texture]
		end
	end

	-- Copy the shared skills to the class skills
	copyTable(faction_db.shared, faction_db.classes[MY_CLASS_UPPER])
end

local function ScanProfessionInfo(index)
	if not index then return end

	local prof_name, _, rank, max_rank = _G.GetProfessionInfo(index)

	-- Update to current values
	MY_PROFESSIONS[prof_name] = MY_PROFESSIONS[prof_name] or {}
	MY_PROFESSIONS[prof_name].rank = rank
	MY_PROFESSIONS[prof_name].max_rank = max_rank

	local prof_ranks = _G.PROFESSION_RANKS

	-- We can infer some training from the max rank
	for i = 1, #prof_ranks do
		local value = prof_ranks[i][1]

		if max_rank >= value then
			-- We know that one
			if type(TITLES[prof_name][value]) ~= 'string' then
				-- Let's keep the calls to GetSpellInfo to a minimum
				local skill_name = _G.GetSpellInfo(TITLES[prof_name][value])
				if skill_name then
					TITLES[prof_name][value] = skill_name
				else
					-- GetSpellInfo didn't return anything
					break
				end
			end
			known_db[prof_name..":"..TITLES[prof_name][value]] = true
		else
			break
		end
	end

end

-- Validate if the toon knowns an ability
local function KnowAbility(ability_name)
	if known_db["ABILITY:"..ability_name] then
		return true
	--else
	--	for name, spell_id in pairs(ABILITY_NAME) do
	--		if strfind(ability_name,name) == 1 then
	--			known_db["ABILITY:"..ability_name] = spell_id
	--			return true
	--		end
	--	end
	end

	return false
end

-- Generate the text that will be used for the data broker bar
local function GetUpdatedBrokerText()
	-- Find the professions that need to be trained
	local total_learnable, total_cost = 0, 0
	local cost_type = db.profile.broker_cost

	for my_prof_name, my_prof in pairs(MY_PROFESSIONS) do
		for prof_name, prof in pairs(faction_db.professions) do
			if prof_name == my_prof_name then
				for skill_name, skill in pairs(prof.skills) do
					--err("pn = %s, sn = %s, mpn = %s",prof_name,skill_name,my_prof_name)
					if not known_db[prof_name..':'..skill_name] and
					   (skill.rank_req == nil or skill.rank_req <= my_prof.rank) and
						(skill.level_req == nil or skill.level_req <= MY_LEVEL) and
						(skill.requires == nil or KnowAbility(skill.requires)) and
						(skill.spell_id == nil or not NEED_ARMOR[skill.spell_id]
						 or NEED_ARMOR[skill.spell_id] == ARMOR_BY_CLASS[MY_CLASS])
					then
						total_learnable = total_learnable + 1
						if cost_type == 'base' then
							total_cost = total_cost + skill.cost
						elseif cost_type == 'cheapest' then
							total_cost = total_cost + GetCheapestCost(skill.cost,skill.valid_rep)
						end
					end
				end
			end
		end
	end

	-- Find the class skills that need to be trained
	if faction_db.classes[MY_CLASS_UPPER] then
		-- When the database is empty because no trainer have been seen yet
		for skill_name, skill in pairs(faction_db.classes[MY_CLASS_UPPER].skills) do
			if not known_db[MY_CLASS_UPPER..':'..skill_name] and
				(skill.level_req == nil or skill.level_req <= MY_LEVEL)  and
				(skill.requires == nil or KnowAbility(skill.requires)) then
				total_learnable = total_learnable + 1
				if cost_type == 'base' then
					total_cost = total_cost + skill.cost
				elseif cost_type == 'cheapest' then
					total_cost = total_cost + GetCheapestCost(skill.cost,skill.valid_rep)
				end
			end
		end
	end

	local text = ""
	if db.profile.broker_learnable then
		text = tonumber(total_learnable)
	end
	if cost_type ~= 'none' and total_cost ~= 0 then
		if text ~= "" then text = text .. ": " end
		text = text .. _G.GetMoneyString(total_cost)
	end

	return text
end

function Syllabus:UpdateSkills()
	if in_combat then
		update_queued = true
		return
	end

	-- Professions related stuff
	wipe(MY_PROFESSIONS)

	local prof1, prof2, arch, fish, cook, firstAid = _G.GetProfessions()

	ScanProfessionInfo(cook)
	ScanProfessionInfo(firstAid)
	ScanProfessionInfo(fish)
	ScanProfessionInfo(arch)
	ScanProfessionInfo(prof1)
	ScanProfessionInfo(prof2)

	-- Class related stuff
	MY_LEVEL = _G.UnitLevel("player")

	-- Update the broker text
	DataObj.text = GetUpdatedBrokerText()
end

local spell_tip = CreateFrame("GameTooltip", "Syllabus_SpellTooltip", UIParent, "GameTooltipTemplate")
local function ShowSkillTooltip(frame, spell_id, button)
	--err("line=%s, arg=%s, button=%s",tostring(cell),tostring(arg),tostring(button))
	if spell_id then
		spell_tip:SetOwner(frame, "ANCHOR_NONE")
		spell_tip:ClearAllPoints()
		--Add TipTac Support
		--if _G.TipTac and _G.TipTac.AddModifiedTip and not spell_tip.tiptac then
		--	_G.TipTac:AddModifiedTip(spell_tip)
		--	spell_tip.tiptac = true
		--end

		spell_tip:SetScale(db.profile.tooltip_scale)
		spell_tip:SetClampedToScreen(true)
		spell_tip:SetSpellByID(spell_id)
		if (frame:GetLeft() + 15) < (spell_tip:GetWidth()) then
			spell_tip:SetPoint("TOPLEFT", frame, "TOPRIGHT", 10, 0)
		else
			spell_tip:SetPoint("TOPRIGHT", frame, "TOPLEFT", -10, 0)
		end
		spell_tip:Show()
	end
end

local function HideSkillTooltip()
	spell_tip:Hide()
end

local yellow_font = CreateFont("LineFont")
yellow_font:CopyFontObject(GameTooltipText)
yellow_font:SetTextColor(1, 215/255, 0) -- Yellow ffd700

local header_font = CreateFont("HeaderFont")
header_font:CopyFontObject(GameTooltipText)
header_font:SetTextColor(1, 1, 1) -- White

local white_font = CreateFont("WhiteFont")
white_font:CopyFontObject(GameTooltipText)
white_font:SetTextColor(1, 1, 1) -- White

local function ShowTooltip(anchor)
	local line, column

	tooltip = QTip:Acquire(MODNAME.."Tooltip", 3, "LEFT", "CENTER", "RIGHT")
	tooltip:SetAutoHideDelay(0.25, anchor)
	tooltip:SetScale(db.profile.tooltip_scale)
	tooltip:SetHeaderFont(header_font)
	tooltip:SetFont(yellow_font)
	tooltip:Clear()
	tooltip:SmartAnchorTo(anchor)

	-- The tables to gather the information to display
	local prof_to_display = AcquireTable()
	local class_to_display = AcquireTable()

	local line, column = tooltip:AddHeader()
	tooltip:SetCell(line, 1,
   			("%s - %s %d %s |cff%s%s|r"):format(MY_NAME, _G.LEVEL, MY_LEVEL, MY_RACE, CLASS_COLORS[MY_CLASS_UPPER], MY_CLASS), header_font, "CENTER", 3)
	tooltip:AddSeparator()

	-- Find the professions that need to be trained
	local total_learnable, total_cost = 0, 0
	local cost_type = db.profile.display_cost
	for my_prof_name, my_prof in pairs(MY_PROFESSIONS) do
		prof_to_display[my_prof_name] = AcquireTable()

		for prof_name, prof in pairs(faction_db.professions) do
			if prof_name == my_prof_name then
				for skill_name, skill in pairs(prof.skills) do
					--err("pn = %s, sn = %s, mpn = %s",prof_name,skill_name,my_prof_name)
					if not known_db[prof_name..':'..skill_name] and
					   (skill.rank_req == nil or skill.rank_req <= my_prof.rank + db.profile.skill_range) and
						(skill.spell_id == nil or not NEED_ARMOR[skill.spell_id] or
						 NEED_ARMOR[skill.spell_id] == ARMOR_BY_CLASS[MY_CLASS]) and
						(skill.requires == nil or not (IS_ABILITY_EXCEPTION_NAME[skill.requires] ~= nil))
					then
					   --err("Adding %s:%s %s",tostring(my_prof_name),tostring(skill_name),tostring(skill.rank_req))
					   local display_skill = AcquireTable()
						display_skill.rank_req = skill.rank_req
						display_skill.level_req = skill.level_req
						display_skill.requires = skill.requires
						display_skill.icon = skill.icon
						display_skill.item_link = skill.item_link
						display_skill.spell_id = skill.spell_id
						display_skill.cost, display_skill.best_rep = GetCheapestCost(skill.cost,skill.valid_rep)
						if cost_type == 'base' then
							-- We reset to base cost if require
							display_skill.cost = skill.cost
						elseif cost_type == 'none' then
							-- We blank the cost if require
							display_skill.cost = nil
						end
						prof_to_display[my_prof_name][skill_name] = display_skill

						-- Running totals
						if (skill.rank_req == nil or skill.rank_req <= my_prof.rank) and
							(skill.level_req == nil or skill.level_req <= MY_LEVEL) and
					   	(skill.requires == nil or KnowAbility(skill.requires)) then
							total_learnable = total_learnable + 1
							total_cost = total_cost + (display_skill.cost or 0)
						end
					end
				end
			end
		end
	end

	-- Find the class skills that need to be trained
	if faction_db.classes[MY_CLASS_UPPER] then
		-- When the database is empty because no trainer have been seen yet
		for skill_name, skill in pairs(faction_db.classes[MY_CLASS_UPPER].skills) do
			if not known_db[MY_CLASS_UPPER..':'..skill_name] and
				(skill.level_req == nil or skill.level_req <= MY_LEVEL + db.profile.level_range) and
				(skill.requires == nil or not (IS_ABILITY_EXCEPTION_NAME[skill.requires] ~= nil))
			then
				local display_skill = AcquireTable()
				display_skill.rank_req = skill.rank_req
				display_skill.level_req = skill.level_req
				display_skill.requires = skill.requires
				display_skill.icon = skill.icon
				display_skill.item_link = skill.item_link
				display_skill.spell_id = skill.spell_id
				display_skill.cost, display_skill.best_rep = GetCheapestCost(skill.cost,skill.valid_rep)
				if cost_type == 'base' then
					-- We reset to base cost if require
					display_skill.cost = skill.cost
				elseif cost_type == 'none' then
					-- We blank the cost if require
					display_skill.cost = nil
				end
				class_to_display[skill_name] = display_skill

				-- Running totals
				if (skill.level_req == nil or skill.level_req <= MY_LEVEL) and
					(skill.requires == nil or KnowAbility(skill.requires)) then
					total_learnable = total_learnable + 1
					total_cost = total_cost + (display_skill.cost or 0)
				end
			end
		end
	end

	-- Display
	-- Sort alpha for the professions
	local sort_prof = AcquireTable()
	for prof_name in pairs(prof_to_display) do
		tinsert(sort_prof,prof_name)
	end
	sort(sort_prof)

	local first_line = true
	for _, prof_name in ipairs(sort_prof) do
		local prof = prof_to_display[prof_name]
		--err("prof_name = %s - %s",prof_name, tostring(prof))

		if not first_line then
			line, column = tooltip:AddSeparator(4,0,0,0,0)
		else
			first_line = nil
		end
		line, column = tooltip:AddHeader()
		tooltip:SetCell(line, 1,
				("%s (%s/%s)"):format(prof_name, MY_PROFESSIONS[prof_name].rank, MY_PROFESSIONS[prof_name].max_rank),
			header_font,
			"LEFT")

		-- Sort profession skills by rank and then alpha
		local sort_skill = AcquireTable()
		for skill_name in pairs(prof) do
			tinsert(sort_skill,skill_name)
			--err("Adding for sort %s:%s",tostring(prof_name),tostring(skill_name))
		end
		sort(sort_skill, function(a,b)
			if prof[a].rank_req ~= prof[b].rank_req then
				return (prof[a].rank_req or 0) < (prof[b].rank_req or 0)
			else
				return a < b
			end
		end)

		for _, skill_name in ipairs(sort_skill) do
			local skill = prof[skill_name]
			local text
			local req_text, separator = "", ""
			local can_train = true

			--err("Display %s:%s %s",tostring(prof_name),tostring(skill_name),tostring(skill.rank_req))
			if skill.rank_req then
				if skill.rank_req > MY_PROFESSIONS[prof_name].rank then
					req_text = ("%s\124cFFFF0000%s\124r"):format(req_text, skill.rank_req)
					can_train = false
				else
					req_text = ("%s\124cFF00FF00%s\124r"):format(req_text, skill.rank_req)
				end
				separator = ", "
			end

			if skill.level_req and skill.level_req > 1 then
				if skill.level_req > MY_LEVEL then
					req_text = ("%s%s\124cFFFF0000%s %s\124r"):format(req_text, separator, _G.LEVEL, skill.level_req)
					can_train = false
				else
					req_text = ("%s%s\124cFF00FF00%s %s\124r"):format(req_text, separator, _G.LEVEL, skill.level_req)
				end
				separator = ", "
			end

			if skill.requires then
				-- Ability requirement are only displayed if missing
				if not KnowAbility(skill.requires)  then
					if db.profile.ability_req then
						local display = SHORT_ABILITY_NAME[skill.requires] or skill.requires
						req_text = ("%s%s\124cFFFF0000%s\124r"):format(req_text, separator, display)
					end
					can_train = false
				end
				--separator = ", "
			end


			if req_text ~= "" then
				req_text = " - "..req_text
			end

			local line_colour = can_train and "\124cFF00FF00" or ""

			--err("Icon: %s",tostring(skill.icon))
			text = ("%s  \124T%s:0:0\124t%s%s"):format(line_colour, skill.icon, skill_name, req_text)

			line, column = tooltip:AddLine()
			tooltip:SetCell(line, 1, text)

			if skill.cost then
				if db.profile.reputation then
					tooltip:SetCell(line, 2, line_colour .. skill.best_rep)
				end
				tooltip:SetCell(line, 3, _G.GetMoneyString(skill.cost), white_font)
			end

			if skill.spell_id then
				tooltip:SetLineScript(line, "OnEnter", ShowSkillTooltip, skill.spell_id)
				tooltip:SetLineScript(line, "OnLeave", HideSkillTooltip)
			end
		end

		ReleaseTable(sort_skill)
	end
	ReleaseTable(sort_prof)

	-- Only display class section if there is something to display	???
		-- Header for the class
		if not first_line then
			line, column = tooltip:AddSeparator(4,0,0,0,0)
		else
			first_line = nil
		end
		line, column = tooltip:AddHeader()
		tooltip:SetCell(line, 1, ("%s (%s %d)"):format(MY_CLASS, _G.LEVEL, MY_LEVEL), header_font, "LEFT")

		-- Sort class skills by level and then alpha
		local sort_class_skill = AcquireTable()
		for skill_name in pairs(class_to_display) do
			tinsert(sort_class_skill,skill_name)
		end
		sort(sort_class_skill, function(a,b)
			if class_to_display[a].level_req ~= class_to_display[b].level_req then
				return (class_to_display[a].level_req or 0) < (class_to_display[b].level_req or 0)
			else
				return a < b
			end
		end)

		for _, skill_name in ipairs(sort_class_skill) do
			local skill = class_to_display[skill_name]
			local text
			local req_text, separator = "", ""
			local can_train = true

			--err("Display %s:%s %s",tostring(prof_name),tostring(skill_name),tostring(skill.rank_req))
			if skill.level_req then
				if skill.level_req > MY_LEVEL then
					req_text = ("%s\124cFFFF0000%s %s\124r"):format(req_text, _G.LEVEL, skill.level_req)
					can_train = false
				else
					req_text = ("%s\124cFF00FF00%s %s\124r"):format(req_text, _G.LEVEL, skill.level_req)
				end
				separator = ", "
			end

			if skill.requires then
				-- Ability requirement are only displayed if missing
				if not KnowAbility(skill.requires) then
					if db.profile.ability_req then
						local display = SHORT_ABILITY_NAME[skill.requires] or skill.requires
						req_text = ("%s%s\124cFFFF0000%s\124r"):format(req_text, separator, display)
					end
					can_train = false
				end
				--separator = ", "
			end
			if req_text ~= "" then
				req_text = " - "..req_text
			end
			--err("Icon: %s",tostring(skill.icon))
			line, column = tooltip:AddLine()

			local line_colour = can_train and "\124cFF00FF00" or ""
				--tooltip::SetLineColor(line, r, g, b)

			text = ("%s  \124T%s:0:0\124t%s%s"):format(line_colour, skill.icon, skill_name, req_text)

			tooltip:SetCell(line, 1, text)

			if skill.cost then
				if db.profile.reputation then
					tooltip:SetCell(line, 2, line_colour .. skill.best_rep)
				end
				tooltip:SetCell(line, 3, _G.GetMoneyString(skill.cost), white_font)
			end

			if skill.spell_id then
				tooltip:SetLineScript(line, "OnEnter", ShowSkillTooltip, skill.spell_id)
				tooltip:SetLineScript(line, "OnLeave", HideSkillTooltip)
			end

		end
		ReleaseTable(sort_class_skill)

	-- Display totals
	if total_learnable ~= 0 and db.profile.total_learnable then
		line, column = tooltip:AddLine()

		tooltip:SetCell(line, 1, " ")
		tooltip:AddSeparator()

		line, column = tooltip:AddLine()

		tooltip:SetCell(line, 1, ("\124cFFFF7F00%s\124r"):format(L['Number of skills ready to train:']))
		tooltip:SetCell(line, 3, tonumber(total_learnable), white_font)
	end

	if total_cost ~= 0 and db.profile.overall_cost then
		line, column = tooltip:AddLine()
		tooltip:SetCell(line, 1, ("\124cFFFF7F00%s\124r"):format(L['Total costs to train the learnable skills:']))
		tooltip:SetCell(line, 3, _G.GetMoneyString(total_cost), white_font)
	end

	-- Upade LDB text
	--DataObj.text = GetUpdatedBrokerText()

	-- Table Clean-up
	ReleaseTable(class_to_display)
	ReleaseTable(prof_to_display)

	-- Adjust the hight of the tooltip so that it fits in the screen
	--tooltip:UpdateScrolling(GetScreenHeight() - 30)
	tooltip:UpdateScrolling()
	tooltip:Show()
end

-------------------------------------------------------------------------------
-- Events
-------------------------------------------------------------------------------
function Syllabus:PLAYER_REGEN_DISABLED()
	in_combat = true
end

function Syllabus:PLAYER_REGEN_ENABLED()
	in_combat = false

	if update_queued then
		update_queued = false
		self:UpdateSkills()
	end
end

function Syllabus:TRADE_SKILL_SHOW()
		-- We don't want to save the information if this isn't our tradeskill list.
	if _G.IsTradeSkillLinked() or _G.IsTradeSkillGuild() then
		return
	end
	DataObj:SetIcon()

	local prof_name = _G.GetTradeSkillLine()
	for i = 1, _G.GetNumTradeSkills() do
		local name, type, _, _, _ = _G.GetTradeSkillInfo(i)

		if name and type ~= "header" then
--			Syllabus:Print("Adding "..name)
			known_db[prof_name..':'..name] = true
		end
	end

	-- Update the broker text
	DataObj.text = GetUpdatedBrokerText()

end

do
	local current_skill = {}	-- Re-usable table for skill insertions.

	local type_filters = {
		["available"] = 0,
		["unavailable"] = 0,
		["used"] = 0,
	}
	local function SavePref()
		-- Save the user's preferences before opening everything.
		for k, v in pairs(type_filters) do
			type_filters[k] = _G.GetTrainerServiceTypeFilter(k) or 0
			_G.SetTrainerServiceTypeFilter(k, 1)
		end

		-- Expend all the headers
		for i = 1, _G.GetNumTrainerServices() do
			local _ , _, category = _G.GetTrainerServiceInfo(i)

			if not category then
				break
			end

			if category == "header" then
				_G.ExpandTrainerSkillLine(i)
			end
		end
	end

	local function RestorePref()
		-- Restore the previous values for the user's type filters.
		for k, v in pairs(type_filters) do
			_G.SetTrainerServiceTypeFilter(k, type_filters[k])
		end
		_G.ClassTrainerFrame.scrollFrame.scrollBar:SetValue(0);
		_G.ClassTrainer_SelectNearestLearnableSkill()
	end

	local DataMiner = CreateFrame("GameTooltip", "SyllabusDataMiner", UIParent, "GameTooltipTemplate")
	DataMiner:SetOwner(UIParent, "ANCHOR_NONE")

	function Syllabus:TRAINER_SHOW()
		_G.GameTooltip_SetDefaultAnchor(DataMiner, _G.UIParent)

		-- Get some information from the trainner tooltip
		DataMiner:SetUnit("npc")

		local trainer_name = _G["SyllabusDataMinerTextLeft1"]:GetText()
		local trainer_line2 = _G["SyllabusDataMinerTextLeft2"]:GetText()
		local trainer_line3 = _G["SyllabusDataMinerTextLeft3"]:GetText()
		-- Some trainers like the DK Class trainers do not have a line 3 i.e. a trainer role
		-- Let's take line 4 for reputation if there is no line 3
		local trainer_rep = _G["SyllabusDataMinerTextLeft4"]:GetText() or trainer_line3
		DataMiner:Hide()

		-- Some NPC have no reputation line but still have 3 lines
		local rep_standing_id = GetReputationStanding(trainer_rep)
		if rep_standing_id == 0 then
			trainer_rep = nil
			rep_standing_id = 4 -- We assume that there are no discount wihtout reputation
		end

		local trainer_role = nil
		local trainer_x, trainer_y = _G.GetPlayerMapPosition("player")
		local trainer_zone = _G.GetRealZoneText()
		local trainer_subzone = _G.GetSubZoneText()

		local is_shared_skill_trainer = false -- identify a shared skill trainer

		if dev then
			print(("Trainer is: %s <%s>, in %s at %.2f, %.2f. Reputation = %s"):format(trainer_name, trainer_line2, trainer_zone, trainer_x * 100, trainer_y * 100,trainer_rep or 'nil'))
		end

		SavePref()

		-- Get the data from the open Trainer Windows
		local trainer_db
		local is_tradeskill = false

		if _G.IsTradeskillTrainer() then
			trainer_db = faction_db.professions
			is_tradeskill = true
		else
			trainer_db = faction_db.classes
		end

		for idx = 1, _G.GetNumTrainerServices() do
			local svc_name, _, svc_type, svc_texture, svc_reqlevel = _G.GetTrainerServiceInfo(idx)

			if svc_name and svc_type and svc_type ~= "header" then
				local level_req	= _G.GetTrainerServiceLevelReq(idx)
				local skill_req, rank_req, _ = _G.GetTrainerServiceSkillReq(idx)

				if rank_req == 0 then
					skill_req = nil
					rank_req = nil
				end
				local num_reqs = _G.GetTrainerServiceNumAbilityReq(idx)
				local ability_reqs, meets_reqs

				if num_reqs > 0 then
					ability_reqs, meets_reqs = _G.GetTrainerServiceAbilityReq(idx, 1)
					-- Save the ability for the toon
					if meets_reqs then
						known_db['ABILITY:'..ability_reqs] = true
					end
				end

				local item_link = _G.GetTrainerServiceItemLink(idx)

				current_skill.item_link = item_link and item_link:match("item:(%d+)") or nil
				current_skill.icon = _G.GetTrainerServiceIcon(idx)
				current_skill.level_req = ((level_req > 0) and level_req) or nil
				--current_skill.description = ((not current_skill.item_link) and GetTrainerServiceDescription(idx)) or nil
				current_skill.cost = GetBaseCost(_G.GetTrainerServiceCost(idx), rep_standing_id)
				current_skill.rank_req = rank_req
				current_skill.requires = ability_reqs
				current_skill.num_reqs = num_reqs
				current_skill.svc_header = _G.GetTrainerServiceSkillLine(idx)

				-- We mine the Tooltip again to get the SpellID
				_G.GameTooltip_SetDefaultAnchor(DataMiner, _G.UIParent)
				DataMiner:SetTrainerService(idx)
				current_skill.spell_id = select(3,DataMiner:GetSpell())
				DataMiner:Hide()

				local svc_header
				if is_tradeskill then
					svc_header = _G.GetTrainerServiceSkillLine(idx)
				else
					svc_header = MY_CLASS_UPPER
				end

				-- Keep the first svc_name for the trainer
				if not trainer_role then
					trainer_role = svc_header
				elseif trainer_role ~= svc_header then
					err("trainer_role changed from %s to %s",trainer_role,svc_header)

				end

				for key, val in pairs(current_skill) do
					trainer_db[svc_header].skills[svc_name][key] = val
				end

				if trainer_rep then
					trainer_db[svc_header].skills[svc_name].valid_rep[trainer_rep] = true
				end
				trainer_db[svc_header].skills[svc_name].trainers[trainer_name] = true

				-- Identify shared skills that are not tradeskills and save the data
				if not is_tradeskill and SHARED_SKILLS[current_skill.spell_id] then
					is_shared_skill_trainer = true

					for key, val in pairs(current_skill) do
						faction_db.shared.skills[svc_name][key] = val
					end

					if trainer_rep then
						faction_db.shared.skills[svc_name].valid_rep[trainer_rep] = true
					end
					faction_db.shared.skills[svc_name].trainers[trainer_name] = true
				end

				if svc_type == "used" and svc_name ~= svc_header then
					known_db[svc_header..':'..svc_name] = true
				end
			end
		end

		local trainer_info = trainer_db[trainer_role].trainers[trainer_name]
		trainer_info.x = trainer_x
		trainer_info.y = trainer_y
		trainer_info.zone = trainer_zone
		trainer_info.subzone = trainer_subzone
		trainer_info.line2 = trainer_line2 or ""
		trainer_info.line3 = trainer_line3 or ""
		trainer_info.rep = trainer_rep

		-- Save the shared skills trainer information
		if is_shared_skill_trainer then
			copyTable(trainer_info, faction_db.shared.trainers[trainer_name])
		end

		RestorePref()
		Syllabus:UpdateSkills()
	end

	function Syllabus:TRAINER_CLOSED()
		--err("TRAINER_CLOSED called")
		-- On close, we just recheck the values on trained skills
		SavePref()

		--local total, total_known = 0, 0
		for idx = 1, _G.GetNumTrainerServices() do
			local header
			local name, subType, category, texture, requiredLevel, topServiceLine = _G.GetTrainerServiceInfo(idx)

			if _G.IsTradeskillTrainer() then
				header = _G.GetTrainerServiceSkillLine(idx)
			else
				header = MY_CLASS_UPPER
			end

			if name and category and category == "used" and name ~= header then
				known_db[header..':'..name] = true
				--total_known = total_known + 1
			end
			--total = total + 1
		end

		--err("TRAINER_CLOSED: total = %s, total_known = %s", total, total_known)

		RestorePref()
		self:UpdateSkills()
	end
end	-- do

-- Clean up the SaveVariables
function Syllabus:PLAYER_LOGOUT()
	-- Debug: for now, don't erase anything
	if 1 then return end

	-- Using GetSpellBookItemInfo
	local i = 1
	local type, spell_id = _G.GetSpellBookItemInfo(i, BOOKTYPE_SPELL)
	while type ~= nil do
		--err("type=%s,spell_id=%s",tostring(type),tostring(spell_id))
		if type == "SPELL" or type == "FLYOUT" then
			local skill_name = (spell_id and _G.GetSpellInfo(spell_id)) or nil
			if skill_name then
				known_db[MY_CLASS_UPPER..":"..skill_name] = nil
			end
		end
		i = i + 1
		type, spell_id = _G.GetSpellBookItemInfo(i, BOOKTYPE_SPELL)
	end
end


-------------------------------------------------------------------------------
-- Initialization functions
-------------------------------------------------------------------------------
function Syllabus:OnInitialize()
	DataObj = LDB:NewDataObject(MODNAME, {
		type	= "data source",
		label	= MODNAME,
		text	= "0",
		icon	= "Interface\\Icons\\INV_Misc_Book_01",
		SetIcon	= function(self, icon)
				  --local num = math.random(15)
				  --self.icon = "Interface\\Icons\\INV_Misc_Book_"..((num < 10) and "0" or "")..num
				  self.icon = icon or "Interface\\Icons\\INV_Misc_Book_01"
				end,
		OnEnter	= function(display, motion)
				  ShowTooltip(display)
				end,
		OnLeave	= function(display, motion)
				end,
		OnClick  = function(display, button, down)
					-- Only when the button is release
					if down then return end

					-- For tests of EasyMenu
					if button ==  "RightButton" then
						_G.InterfaceOptionsFrame_OpenToCategory(MODNAME)
					end
				end,
	})

	-- Defaults describe the data structure
	local defaults = {
		global = {
			Alliance = {
				classes = {
					-- Class Name
					['*'] = {
						trainers = {
							-- Trainer name
							['*'] = {},
						},
						skills = {
							-- Skill name
							['*'] = {
								valid_rep = {},
								trainers = {},
							},
						},
					},
				},
				professions = {
					-- Profession Name
					['*'] = {
						trainers = {
							-- Trainer name
							['*'] = {},
						},
						skills = {
							-- Skill name
							['*'] = {
								valid_rep = {},
								trainers = {},
							},
						},
					},
				},
				shared = {
					trainers = {
						-- Trainer name
						['*'] = {},
					},
					skills = {
						-- Skill name
						['*'] = {
							valid_rep = {},
							trainers = {},
						},
					},
				},
			},
			Horde = {
				classes = {
					-- Class Name
					['*'] = {
						trainers = {
							-- Trainer name
							['*'] = {},
						},
						skills = {
							-- Skill name
							['*'] = {
								valid_rep = {},
								trainers = {},
							},
						},
					},
				},
				professions = {
					-- Profession Name
					['*'] = {
						trainers = {
							-- Trainer name
							['*'] = {},
						},
						skills = {
							-- Skill name
							['*'] = {
								valid_rep = {},
								trainers = {},
							},
						},
					},
				},
				shared = {
					trainers = {
						-- Trainer name
						['*'] = {},
					},
					skills = {
						-- Skill name
						['*'] = {
							valid_rep = {},
							trainers = {},
						},
					},
				},
			},
			hidden = {},
			known = {
				-- Realm:Character Name
				['*'] = {},
			},
		},
		char = {
			known = {
				-- Class or profession name:Skill Name
				['*'] = nil,
			},
		},
		profile = {
			-- Configuration data
			-- Data Broker display
			broker_learnable = true,
			-- cost = none, base or cheapest
			broker_cost = 'cheapest',

			-- Main Tooltip dispaly
			reputation = false,
			display_cost = 'cheapest',
			total_learnable = true,
			overall_cost = true,
			level_range = 4,
			skill_range = 10,
			ability_req = true,
			tooltip_scale = 1,

			not_trainable = false,
			requirements = true,
			skill_count = true,
			trainable = true,
			trainable_future = true,
			trainer_names = true,
			skill_tooltip = true,
		}
	}
	db = _G.LibStub("AceDB-3.0"):New(MODNAME.."DB", defaults)
	Syllabus.db = db
	faction_db = db.global[MY_FACTION]
	known_db = db.char.known

	self:SetupOptions()
	self.OnInitialize = nil
end

function Syllabus:OnEnable()
	self:InitKnownspells()
	self:UpdateSkills()
	self:RegisterEvent("PLAYER_LEVEL_UP", "UpdateSkills")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("SKILL_LINES_CHANGED", "UpdateSkills")
	self:RegisterEvent("TRADE_SKILL_SHOW")
	self:RegisterEvent("TRAINER_SHOW")
	self:RegisterEvent("TRAINER_CLOSED")
	self:RegisterEvent("PLAYER_LOGOUT")
end

function Syllabus:OnDisable()
	self:UnregisterEvent("PLAYER_LEVEL_UP")
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	self:UnregisterEvent("SKILL_LINES_CHANGED")
	self:UnregisterEvent("TRADE_SKILL_SHOW")
	self:UnregisterEvent("TRAINER_SHOW")
	self:UnregisterEvent("TRAINER_CLOSED")
end
